#!/usr/bin/perl
#
#
# Backup the contents of a directory to another directory.
#
#

my $CMDNAME="bakdir";
my $VERSION="3.2";
my $COPYRIGHT="copyright (C) 2010-2024 Frits Wiarda";

######################################################################
#
# Help text.
#
######################################################################

my $HELP=	"Backup the contents of a directory to another directory.\n" .
		"\n" .
		"Usage: bakdir [-a] [-d] sourcedir targetdir [logfile]\n" .
		"\n" .
		"Unless the -a option is specified, files with a name\n" .
		"starting with a dot are ignored\n" .
		"\n" .
		"If the -d option is specified, files in targetdir that do\n" .
		"not exist in sourcedir are deleted.\n";
my $USAGE="Usage: bakdir [-a] [-d] sourcedir targetdir [logfile]\n";

######################################################################
#
# Set parameters.
#
######################################################################

my $SLASH="/";
my $QUOTE="'";
my $ESCAPED_SINGLE_QUOTE="'\\''";
my $MKDIR="mkdir";
my $RMDIR="rm -rf";
my $CP="cp -pf";
my $RM="rm -f";

######################################################################
#
# Global variables.
#
######################################################################

my $leftdir;
my $rightdir;
my $logfile;
my $option_all;
my $option_delete;


######################################################################
#
# Print help message when the option "-?", "/?" or "--help" is 
# specified.
#
######################################################################

# Print help message.
if($ARGV[0] eq "/?" or $ARGV[0] eq "-?" or $ARGV[0] eq "--help")
{
	print "$CMDNAME $VERSION   $COPYRIGHT\n\n$HELP";
	exit;
}

######################################################################
#
# Start execution of the actual code.
#
######################################################################

my $arg_counter;
my $non_option_arg_counter;

# Print usage message when command line parameters are not OK.
# command line.
for($arg_counter=0; $ARGV[$arg_counter] ne ""; $arg_counter++)
{
	if($ARGV[$arg_counter] =~ /^-/ )
	{
		if($ARGV[$arg_counter] =~ /[^-ad]/ )
		{
			die "$USAGE";
		}
	}
	else
	{
		$non_option_arg_counter++;
	}
}
if($non_option_arg_counter<2 || $non_option_arg_counter>3)
{
	die "$USAGE";
}

# Evaluate the command line options.
$option_all=0;
$option_delete=0;
for($arg_counter=0; $ARGV[$arg_counter] ne ""; $arg_counter++)
{
	if($ARGV[$arg_counter] =~ /^-.*a/ )
	{
		$option_all=1;
	}
	if($ARGV[$arg_counter] =~ /^-.*d/ )
	{
		$option_delete=1;
	}
}

# Find first non-option argument.
for($arg_counter=0; $ARGV[$arg_counter] ne ""; $arg_counter++)
{
	if( not $ARGV[$arg_counter] =~ /^-/ )
	{
		last;
	}
}

# Copy leftdir from command line, test if it exist,
# change backslashes into slashes and append a slash if 
# it is not ending with a slash yet.
$leftdir=$ARGV[$arg_counter];
if(not -d $leftdir)
{
	die "$CMDNAME: '$leftdir' is not a directory.\n";
}
$leftdir=~s/\\/\//g;
if(not $leftdir=~/\/$/)
{
	$leftdir.="/";
}

# Find second non-option argument.
for($arg_counter++; $ARGV[$arg_counter] ne ""; $arg_counter++)
{
	if( not $ARGV[$arg_counter] =~ /^-/ )
	{
		last;
	}
}

# Copy rightdir from command line, test if it exist,
# change backslashes into slashes and append a slash if 
# it is not ending with a slash yet.
$rightdir=$ARGV[$arg_counter];
if(not -d $rightdir)
{
	die "$CMDNAME: '$rightdir' is not a directory.\n";
}
$rightdir=~s/\\/\//g;
if(not $rightdir=~/\/$/)
{
	$rightdir.="/";
}

# Get the name of the logfile.
if($non_option_arg_counter eq 3)
{

	# Find third non-option argument.
	for($arg_counter++; $ARGV[$arg_counter] ne ""; $arg_counter++)
	{
		if( $ARGV[$arg_counter] =~ /^[^-]/ )
		{
			last;
		}
	}

	# Copy the found argument.
	$logfile=$ARGV[$arg_counter];
}

# Or set it to "" if none has been specified.
else
{
	$logfile="";
}

# Create an empty log file (when one has been specified).
if($logfile ne "")
{
	local *FILE;
	open(FILE,">$logfile") or 
		die "$CMDNAME: Cannot open file '$logfile'.\n";
	close(FILE);
}

# Walk through the directory tree.
walk_tree("");

# Ready.
exit 0;


######################################################################
#
# Subroutine to walk through the directory tree.
#
######################################################################

sub walk_tree()
{
	my $subdir=$_[0];

	local *DIR;
	my @leftdir_contents;
	my @rightdir_contents;
	my @sorted_dir_contents;
	my @relevant_dir_contents;
	my $dir_entry;

	my $left_size;
	my $left_time;
	my $right_size;
	my $right_time;

	my $dev;
	my $ino;
	my $mode;
	my $nlink;
	my $uid;
	my $gid;
	my $rdev;
	my $atime;
	my $ctime;
	my $blksize;
	my $blocks;

	my $i;
	my $j;

	# Read directory to the left.
	opendir(DIR,"$leftdir$subdir") or 
		die("$CMDNAME: Cannot open directory '$leftdir$subdir'. Backup aborted.\n");
	@leftdir_contents=readdir(DIR);
	closedir(DIR);

	# Read directory to the right.
	opendir(DIR,"$rightdir$subdir") or 
		die("$CMDNAME: Cannot open directory '$rightdir$subdir'. Backup aborted.\n");
	@rightdir_contents=readdir(DIR);
	closedir(DIR);

	# Concatenate directory contents and sort them alphabetically.
	@sorted_dir_contents=sort (@leftdir_contents,@rightdir_contents);

	# Remove double entries, "." and "..", and when the "-a" 
	# option is not specified at the command line, all other 
	# entries starting with ".".
	for($i=0,$j=0;$i<=$#sorted_dir_contents;$i++)
	{
		if($i>0 and $sorted_dir_contents[$i] eq 
			$sorted_dir_contents[$i-1])
		{
			next;
		}
		if($sorted_dir_contents[$i] eq ".")
		{
			next;
		}
		if($sorted_dir_contents[$i] eq "..")
		{
			next;
		}
		if($option_all eq 0 and $sorted_dir_contents[$i] =~ /^\./)
		{
			next;
		}
		$relevant_dir_contents[$j]=$sorted_dir_contents[$i];
		$j++;
	}

	# Now loop through the array relevant_dir_contents. Do the
	# removal part of the synchronization first. Remove before
	# copy ensures correct functioning on filesystems that
	# consider lowercase and uppercase names to be identical. 
	if($option_delete){
		foreach $dir_entry (@relevant_dir_contents)
		{
			# Determine full filename to the left and to the right.
			my $left="$leftdir$subdir$dir_entry";
			my $right="$rightdir$subdir$dir_entry";	

			# Both left and right are files.
			if(-f $left and -f $right)
			{
			}

			# Both left and right are directories.		
			elsif(-d $left and -d $right)
			{
			}

			# Left is a file and right does not exist.
			elsif(-f $left and not -e $right)
			{
			}

			# Left does not exist and right is a file.
			elsif(not -e $left and -f $right)
			{
	
				# Check that parent directory of deleted 
				# file is still readable. This is done to 
				# prevent erroneous deletes when a network
				# drive gets disconnected while backing up.
				if(not -d "$leftdir$subdir")
				{
					die "$CMDNAME: Directory '$leftdir$subdir' is not readable. Backup aborted.\n";
				}

				# Delete file to the right.
				execute_command("$RM \"$right\"");
			}

			# Left is a directory and right does not exist.
			elsif(-d $left and not -e $right)
			{
			}

			# Left does not exist and right is a directory.
			elsif(not -e $left and -d $right)
			{
				# Check that parent directory of deleted 
				# file is still readable. This is done to 
				# prevent erroneous deletes when a network
				# drive gets disconnected while backing up.
				if(not -d "$leftdir$subdir")
				{
					die "$CMDNAME: Directory '$leftdir$subdir' is not readable. Backup aborted.\n";
				}

				# Delete directory (tree) to the right.
				execute_command("$RMDIR \"$right\"");
			}

			# Left is a directory and right is a file.
			elsif(-d $left and -f $right)
			{
			}

			# Left is a file and right is a directory.
			elsif(-f $left and -d $right)
			{
			}		

			# The left entry is neither a file nor a directory.
			# E.g. a device file. We do not support that.
			elsif(-e $left and not -d $left and not -f $left)
			{
				die "$CMDNAME: Unrecognized file type for '$left'. Backup aborted.\n"; 
			}

			# The right entry is neither a file nor a directory.
			# E.g. a device file. We do not support that.
			elsif(-e $right and not -d $right and not -f $right)
			{
				die "$CMDNAME: Unrecognized file type for '$right'. Backup aborted.\n"; 
			}

			# We should not be able to get here.
			else
			{
				die "$CMDNAME: Panic. Backup aborted.\n"; 
			}
		}
	}

	# Now loop through the array relevant_dir_contents and do
	# the backup.
	foreach $dir_entry (@relevant_dir_contents)
	{
		# Determine full filename to the left and to the right.
		my $left="$leftdir$subdir$dir_entry";
		my $right="$rightdir$subdir$dir_entry";	

		# Both left and right are files.
		if(-f $left and -f $right)
		{
			# Get the last modification time stamps and file sizes.
			($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,
				$left_size,$atime,$left_time,$ctime,$blksize,$blocks)
				=stat($left);
			($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,
				$right_size,$atime,$right_time,$ctime,$blksize,$blocks)
				=stat($right);

			# Check if left and right do not have equal
			# time stamps.
			if( gt_time($left_time,$right_time,
				$left_size eq $right_size) or
				gt_time($right_time,$left_time,
				$left_size eq $right_size) )
			{
				# Update file to the right.
				execute_command("$CP \"$left\" \"$right\"");
			}

			# Otherwise files are in sync.
			else
			{
			}

		}

		# Both left and right are directories.		
		elsif(-d $left and -d $right)
		{
			# Process everything inside the directory.
			walk_tree("$subdir$dir_entry/");
		}

		# Left is a file and right does not exist.
		elsif(-f $left and not -e $right)
		{
			# Copy file to the right.
			execute_command("$CP \"$left\" \"$right\"");
		}

		# Left does not exist and right is a file.
		elsif(not -e $left and -f $right)
		{
		}

		# Left is a directory and right does not exist.
		elsif(-d $left and not -e $right)
		{
			# Create an empty directory to the right.
			execute_command("$MKDIR \"$right\"");
	
			# Process everything inside the directory.
			walk_tree("$subdir$dir_entry/");
		}

		# Left does not exist and right is a directory.
		elsif(not -e $left and -d $right)
		{
		}

		# Left is a directory and right is a file.
		elsif(-d $left and -f $right)
		{
			# Replace file to the right with an empty
			# directory.
			execute_command("$RM \"$right\"");
			execute_command("$MKDIR \"$right\"");

			# Now process everything inside the directory.
			walk_tree("$subdir$dir_entry/");
		}

		# Left is a file and right is a directory.
		elsif(-f $left and -d $right)
		{

			# Replace directory (tree) to the right 
			# with file to the left.							
			execute_command("$RMDIR \"$right\"");
			execute_command("$CP \"$left\" \"$right\"");
		}		

		# Both entrys do not exist. This is because in the
		# entry has been removed before.
		elsif(not -e $left and not -e)
		{
		}

		# The left entry is neither a file nor a directory.
		# E.g. a device file. We do not support that.
		elsif(-e $left and not -d $left and not -f $left)
		{
			die "$CMDNAME: Unrecognized file type for '$left'. Backup aborted.\n"; 
		}

		# The right entry is neither a file nor a directory.
		# E.g. a device file. We do not support that.
		elsif(-e $right and not -d $right and not -f $right)
		{
			die "$CMDNAME: Unrecognized file type for '$right'. Backup aborted.\n"; 
		}

		# We should not be able to get here.
		else
		{
			die "$CMDNAME: Panic. Backup aborted.\n"; 
		}
	}
}

######################################################################
#
# Subroutine to execute a command and log it as well.
#
######################################################################

sub execute_command()
{
	my $command=$_[0];

        # Replace single quotes with escaped single quotes.
        $command=~s/'/$ESCAPED_SINGLE_QUOTE/g;

	# If the system slash is not a forward slash,
	# replace it.
	if($SLASH ne "/")
	{
		$command=~s/\//$SLASH/g;
	}

	# If the system quote is not a double quote,
	# replace it.
	if($QUOTE ne "\"")
	{
		$command=~s/"/$QUOTE/g;
	}

	# Log the command to be executed to the standard output.
	print "$command\n";

	# If a logfile has been specified on the command line,
	# log the command to be executed to that file as well.
	if($logfile ne "")
	{
		local *FILE;
		open(FILE,">>$logfile") or 
			die "$CMDNAME: Cannot open file '$logfile'.\n";
		print FILE "$command\n";
		close(FILE);
	}

	# Execute the command. Exit this program if the command 
	# exits with an error.
	if(system($command) ne 0)
	{
		die "$CMDNAME: Backup aborted.\n";
	}
}


######################################################################
#
# Check if file time t1 is greater t2. If ignore_one_hour is true,
# a time difference of 3600 s is considered to be no time difference.
# Accuracy of time is +/- 3 s.
#
######################################################################

sub gt_time()
{
	my $t1=$_[0];
	my $t2=$_[1];
	my $ignore_one_hour=$_[2];

	# Time differences of one hour shall be ignored.
	if($ignore_one_hour)
	{
		# Time t1 is later then time t2.
		if($t1-$t2 > 3 and 
			($t1-$t2 > 3603 or $t1-$t2 < 3597 ) )
		{
			return 1;
		}

		# Time t1 is not later then time t2.
		else
		{
			return 0
		}
	}

	# Time differences of one hour shall not be ignored.
	else
	{
		# Time t1 is later then time t2.
		if($t1-$t2 > 3)
		{
			return 1;
		}

		# Time t1 is not later then time t2.
		else
		{
			return 0;
		}
	}
}
