/**
*
* Backup the contents of a directory to another directory.
*
**/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <windows.h>
#include <sys/types.h>
#include <sys/stat.h>

#define CMDNAME 		"bakdir"
#define VERSION 		"3.2"
#define COPYRIGHT		"copyright (C) 2010-2024 Frits Wiarda"
#define HELP			"Backup the contents of a directory to another directory.\n"	\
				"\n"					\
				"Usage: bakdir [-a] [-d] sourcedir targetdir [logfile]\n"\
				"\n"\
				"Unless the -a option is specified, files with a name\n"\
				"starting with a dot, hidden files and system file are ignored.\n"\
				"\n"\
				"If the -d option is specified, files in targetdir that do\n"\
				"not exist in sourcedir are deleted.\n"
#define USAGE			"Usage: bakdir [-a] [-d] sourcedir targetdir [logfile]\n"

#define SLASH			'\\'
#define QUOTE			'"'
#define ESCAPED_SINGLE_QUOTE	"'"
#define MKDIR			"mkdir"
#define RMDIR			"rm -rf"
#define CP			"cp -pf"
#define RM			"rm -f"

#define DIR_SIZE 		9192
#define FILENAME_SIZE		512
#define PARAMETER_SIZE		512

char leftdir[PARAMETER_SIZE];
char rightdir[PARAMETER_SIZE];
char *logfile;
int option_all;
int option_delete;



void walk_tree(char *);
void execute_command(char *);
int check_database(char *,char *,char *);
int gt_time(FILETIME,FILETIME,int);
int dir_exists(char *);



int main(int argc, char *argv[])
{

	FILE *fp;

	int arg_counter,non_option_arg_counter;
	int i;

	/* If first and only argument is "-?", "/?" or "--help",
	print copyright notice and help information. */
	if(argc==2 && (argv[1][0]=='-'||argv[1][0]=='/') && 
		argv[1][1]=='?' && argv[1][2]=='\0' || argc==2 && 
		argv[1][0]=='-' && argv[1][1]=='-' && 
		argv[1][2]=='h' && argv[1][3]=='e' && 
		argv[1][4]=='l' && argv[1][5]=='p' && 
		argv[1][6]=='\0')
	{
		fprintf(stdout,"%s %s   %s\n\n%s",
			CMDNAME,VERSION,COPYRIGHT,HELP);
		return 0;
	}

	/* Print usage message if parameters are not OK. */
	non_option_arg_counter=0;
	for(arg_counter=1; arg_counter<argc; arg_counter++)
	{
		if(argv[arg_counter][0]=='-')
		{
			if( strcmp(argv[arg_counter],"-a")!=0 &&
				strcmp(argv[arg_counter],"-d")!=0 &&
				strcmp(argv[arg_counter],"-ad")!=0 &&
				strcmp(argv[arg_counter],"-da")!=0 )
			{
				fprintf(stderr,"%s",USAGE);
				exit(1);
			}
		}
		else non_option_arg_counter++;
	}
	if(non_option_arg_counter<2 || non_option_arg_counter>3)
	{
		fprintf(stderr,"%s",USAGE);
		exit(1);
	}

	/* Evaluate the command line options. */
	option_all=0;
	option_delete=0;
	for(arg_counter=1; arg_counter<argc; arg_counter++)
	{
		if(argv[arg_counter][0]=='-')
		{
			for(i=1; argv[arg_counter][i]!='\0'; i++)
			{
				if(argv[arg_counter][i]=='a')
				{
					option_all=1;
				}
				else if(argv[arg_counter][i]=='d')
				{
					option_delete=1;
				}
			}
		}
	}

	/* Find the first non-option argument. */
	for(arg_counter=1; arg_counter<argc; arg_counter++)
	{
		if(argv[arg_counter][0]!='-')break;
	}

	/* Copy leftdir from command line, test if it exist,
	change backslashes into slashes and append a slash if 
	it is not ending with a slash yet. */
	if(!dir_exists(argv[arg_counter]))
	{
		fprintf(stderr,"%s: '%s' is not a directory.\n",CMDNAME,argv[arg_counter]);
		exit(1);
	}
	for(i=0;argv[arg_counter][i]!='\0' && i<PARAMETER_SIZE-2;i++)
	{
		leftdir[i]=argv[arg_counter][i];
		if(leftdir[i]=='\\')leftdir[i]='/';
	}
	if(leftdir[i-1]!='/')leftdir[i++]='/';
	leftdir[i]='\0';

	/* Find the second non-option argument. */
	for(arg_counter++; arg_counter<argc; arg_counter++)
	{
		if(argv[arg_counter][0]!='-')break;
	}

	/* Copy rightdir from command line, test if it exist,
	change backslashes into slashes and append a slash if 
	it is not ending with a slash yet. */
	if(!dir_exists(argv[arg_counter]))
	{
		fprintf(stderr,
			"%s: '%s' is not a directory.\n",CMDNAME,argv[arg_counter]);
		exit(1);
	}
	for(i=0;argv[arg_counter][i]!='\0' && i<PARAMETER_SIZE-2;i++)
	{
		rightdir[i]=argv[arg_counter][i];
		if(rightdir[i]=='\\')rightdir[i]='/';
	}
	if(rightdir[i-1]!='/')rightdir[i++]='/';
	rightdir[i]='\0';

	/*  Copy the optional logfile from the command line and create an 
	empty one (when it has been specified). */
	if(non_option_arg_counter==3)
	{
		/* Find the third non-option argument. */
		for(arg_counter++; arg_counter<argc; arg_counter++)
		{
			if(argv[arg_counter][0]!='-')break;
		}

		logfile=argv[arg_counter];
		fp=fopen(logfile,"wt");
		if(fp==NULL)
		{
			fprintf(stderr,"%s: Cannot open file '%s'.\n",
				CMDNAME,logfile);
			exit(1);
		}
		fclose(fp);
	}
	else
	{
		logfile=NULL;
	}

	/* Walk through the directory tree. */
	walk_tree("");

	/* Ready. */
	return 0;

}



/***
*
* Walk through the directory tree.
*
***/

void walk_tree(char *subdir)
{

	struct dir_contents {
		char *name;
		char left_type;
		FILETIME left_time;
		long int left_size;
		char right_type;
		FILETIME right_time;
		long int right_size;
	} *dir_contents, tmp_dir_contents;

	int dir_contents_sorted;

	char *string;

	int i;
	int j;

	WIN32_FIND_DATA fileinfo;
	HANDLE h;

	/* Allocate memory for temporary strings. The maximum size needed
	is a command plus spaces and quotes plus a left pathname plus a
	right pathname. */
	string=malloc( 80 + 
		strlen(leftdir) + strlen(subdir) + FILENAME_SIZE + 
		strlen(rightdir) + strlen(subdir) + FILENAME_SIZE );
	if(string==NULL)
	{
		fprintf(stderr,"%s: Out of memory. Backup aborted.\n",
			CMDNAME);
		exit(1);
	}

	/* Allocate memory for directory contents. */
	dir_contents=malloc(sizeof(struct dir_contents)*DIR_SIZE);
	if(dir_contents==NULL)
	{
		fprintf(stderr,"%s: Out of memory. Backup aborted.\n",
			CMDNAME);
		exit(1);
	}

	/* Set the terminating NULL to dir_contents. */
	dir_contents[0].name=NULL;

	/* Read directory to the left, and copy the information into
	dir_contents. */
	sprintf(string,"%s%s*",leftdir,subdir);
	h=FindFirstFileA((LPCSTR)string,&fileinfo);
	if(h!=INVALID_HANDLE_VALUE)for(i=0;;)
	{
		/* We always skip "." and "..". */
		if(strcmp(fileinfo.cFileName,".")==0 || 
			strcmp(fileinfo.cFileName,"..")==0 )
		{
			if(FindNextFileA(h,&fileinfo)==0)break;
			else continue;
		}

		/* We skip hidden and system files, and files with a name
		starting with a dot, unless the "-a" option has been 
		specified. */
		if(!option_all)
		{
			/* Hidden and system files. */
			if(fileinfo.dwFileAttributes & 
				FILE_ATTRIBUTE_HIDDEN ||
				fileinfo.dwFileAttributes & 
				FILE_ATTRIBUTE_SYSTEM)
			{
				if(FindNextFileA(h,&fileinfo)==0)break;
				else continue;
			}

			/* Files with a name starting with a dot. */
			if(fileinfo.cFileName[0]=='.')
			{
				if(FindNextFileA(h,&fileinfo)==0)break;
				else continue;
			}
		}

		/* Add entry to dir_contents. */
		dir_contents[i].name=malloc(strlen(fileinfo.cFileName)+1);
		if(dir_contents[i].name==NULL)
		{
			fprintf(stderr,
				"%s: Out of memory. Backup aborted.\n",
				CMDNAME);
			exit(1);
		}
		strcpy(dir_contents[i].name,fileinfo.cFileName);
		if(fileinfo.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
		{
			dir_contents[i].left_type='d';
		}
		else 
		{
			dir_contents[i].left_type='f';
		}
		dir_contents[i].left_time=fileinfo.ftLastWriteTime;
      		dir_contents[i].left_size=fileinfo.nFileSizeHigh * MAXDWORD +
			fileinfo.nFileSizeLow;
		dir_contents[i].right_type='n';

		/* Increase counter and add the terminating NULL. */
		i++;
		dir_contents[i].name=NULL;

		/* Get next entry. */
		if(FindNextFileA(h,&fileinfo)==0)break;

		/* Check that there is space left in the array dir_contents. */
		if(i>=DIR_SIZE-1)
		{
			fprintf(stderr,
				"%s: To many entries in directory '%s%s'. Backup aborted.\n",
				CMDNAME,leftdir,subdir);
			exit(1);
		}
	}
	if(h!=INVALID_HANDLE_VALUE)FindClose(h);

	/* Read directory to the right, and copy the information into
	dir_contents. */
	sprintf(string,"%s%s*",rightdir,subdir);
	h=FindFirstFileA((LPCSTR)string,&fileinfo);
	if(h!=INVALID_HANDLE_VALUE)for(;;)
	{

		/* We always skip "." and "..". */
		if(strcmp(fileinfo.cFileName,".")==0 || 
			strcmp(fileinfo.cFileName,"..")==0 )
		{
			if(FindNextFileA(h,&fileinfo)==0)break;
			else continue;
		}

		/* We skip hidden and system files, and files with a name
		starting with a dot, unless the "-a" option has been 
		specified. */
		if(!option_all)
		{
			/* Hidden and system files. */
			if(fileinfo.dwFileAttributes & 
				FILE_ATTRIBUTE_HIDDEN ||
				fileinfo.dwFileAttributes & 
				FILE_ATTRIBUTE_SYSTEM)
			{
				if(FindNextFileA(h,&fileinfo)==0)break;
				else continue;
			}

			/* Files with a name starting with a dot. */
			if(fileinfo.cFileName[0]=='.')
			{
				if(FindNextFileA(h,&fileinfo)==0)break;
				else continue;
			}
		}

		/* Try to find filename in dir_contents structure. */
		for(j=0;dir_contents[j].name!=NULL;j++)
		{
			if(strcmp(dir_contents[j].name,
				fileinfo.cFileName)==0)break;
		}

		/* If filename found we only have to add the file data to
		the existing entry. */		
		if(dir_contents[j].name!=NULL)
		{
			/* Add file data to existing entry. */
			if(fileinfo.dwFileAttributes & 
				FILE_ATTRIBUTE_DIRECTORY)
			{
				dir_contents[j].right_type='d';
			}
			else 
			{
				dir_contents[j].right_type='f';
			}			
			dir_contents[j].right_time=fileinfo.ftLastWriteTime;
			dir_contents[j].right_size=
				fileinfo.nFileSizeHigh * MAXDWORD +
				fileinfo.nFileSizeLow;
		}

		/* Otherwise we have to create a new entry. */
		else
		{
			/* Add entry to dir_contents. */
			dir_contents[i].name=
				malloc(strlen(fileinfo.cFileName)+1);
			if(dir_contents[i].name==NULL)
			{
				fprintf(stderr,
					"%s: Out of memory. Backup aborted.\n",
					CMDNAME);
				exit(1);
			}
			strcpy(dir_contents[i].name,fileinfo.cFileName);
			dir_contents[i].left_type='n';
			if(fileinfo.dwFileAttributes & 
				FILE_ATTRIBUTE_DIRECTORY)
			{
				dir_contents[i].right_type='d';
			}
			else 
			{
				dir_contents[i].right_type='f';
			}
			dir_contents[i].right_time=fileinfo.ftLastWriteTime;
			dir_contents[i].right_size=
				fileinfo.nFileSizeHigh * MAXDWORD +
				fileinfo.nFileSizeLow;

			/* Increase counter and add the terminating NULL. */
			i++;
			dir_contents[i].name=NULL;
		}

		/* Get next entry. */
		if(FindNextFileA(h,&fileinfo)==0)break;

		/* Check that there is space left in the array dir_contents. */
		if(i>=DIR_SIZE-1)
		{
			fprintf(stderr,
				"%s: To many entries in directory '%s%s'. Backup aborted.\n",
				CMDNAME,leftdir,subdir);
			exit(1);
		}
	}
	if(h!=INVALID_HANDLE_VALUE)FindClose(h);

	/* Sort the array dir_contents when there is more then
	one entry in it. */
	if(dir_contents[0].name!=NULL && dir_contents[1].name!=NULL)
	{
		do
		{
			dir_contents_sorted=1;
			for(i=1;dir_contents[i].name!=NULL;i++)
			{
				if(strcmp(dir_contents[i-1].name,
					dir_contents[i].name)>0)
				{
					dir_contents_sorted=0;
						tmp_dir_contents=
							dir_contents[i-1];
						dir_contents[i-1]=
							dir_contents[i];
						dir_contents[i]=
							tmp_dir_contents;
				}
			}
		}
		while(!dir_contents_sorted);
	}

	/* Now loop through the array dir_contents and do the
	removal. This is only done if the "-d" option has been
	specified. Remove before copy ensures correct functioning
	on filesystems that consider lowercase and uppercase names 
	to be identical. */
	if(option_delete)for(i=0;dir_contents[i].name!=NULL;i++)
	{

		/* Both left and right are files. */
		if(dir_contents[i].left_type=='f' && 
			dir_contents[i].right_type=='f')
		{
		}

		/* Both left and right are directories.	*/
		else if(dir_contents[i].left_type=='d' && 
			dir_contents[i].right_type=='d')
		{
		}

		/* Left is a file and right does not exist. */
		else if(dir_contents[i].left_type=='f' && 
			dir_contents[i].right_type=='n')
		{
		}

		/* Left does not exist and right is a file. */
		else if(dir_contents[i].left_type=='n' && 
			dir_contents[i].right_type=='f')
		{
	
			/* Check that parent directory of deleted 
			file is still readable. This is done to 
			prevent a whole tree being deleted if a 
			network drive gets disconnected while 
			backup. */
			sprintf(string,"%s%s",leftdir,subdir);
			if(!dir_exists(string))
			{
				fprintf(stderr,
					"%s: Directory '%s' is not readable. Backup aborted.\n",
					CMDNAME,string);
				exit(1);
			}

			/* Delete file to the right. */
			sprintf(string,"%s \"%s%s%s\"",RM,
				rightdir,subdir,dir_contents[i].name);
			execute_command(string);

		}

		/* Left is a directory and right does not exist. */
		else if(dir_contents[i].left_type=='d' && 
			dir_contents[i].right_type=='n')
		{
		}

		/* Left does not exist and right is a directory. */
		else if(dir_contents[i].left_type=='n' && 
			dir_contents[i].right_type=='d')
		{

			/* Check that parent directory of deleted 
			file is still readable. This is done to 
			prevent a whole tree being deleted if a 
			network drive gets disconnected while 
			backup. */
			sprintf(string,"%s%s",leftdir,subdir);
			if(!dir_exists(string))
			{
				fprintf(stderr,
					"%s: Directory '%s' is not readable. Backup aborted.\n",
					CMDNAME,string);
				exit(1);
			}

			/* Delete directory (tree) to the right. */	
			sprintf(string,"%s \"%s%s%s\"",RMDIR,
				rightdir,subdir,dir_contents[i].name);
			execute_command(string);
		}

		/* Left is a directory and right is a file. */
		else if(dir_contents[i].left_type=='d' && 
			dir_contents[i].right_type=='f')
		{
		}

		/* Left is a file and right is a directory. */
		else if(dir_contents[i].left_type=='f' && 
			dir_contents[i].right_type=='d')
		{
		}		

		/* We should not be able to get here. */
		else
		{
			fprintf(stderr,
				"%s: Panic. Backup aborted.\n",
				CMDNAME);
				exit(1);
		}
	}

	/* Now loop through the array dir_contents and do the
	backup. */
	for(i=0;dir_contents[i].name!=NULL;i++)
	{

		/* Both left and right are files. */
		if(dir_contents[i].left_type=='f' && 
			dir_contents[i].right_type=='f')
		{

			/* Check if left and right do not have equal
			time stamps. */
			if( gt_time(dir_contents[i].left_time,
				dir_contents[i].right_time,
				dir_contents[i].left_size==
				dir_contents[i].right_size) ||
				gt_time(dir_contents[i].right_time,
				dir_contents[i].left_time,
				dir_contents[i].left_size==
				dir_contents[i].right_size) )
			{
				/* Update file to the right. */
				sprintf(string,"%s \"%s%s%s\" \"%s%s%s\"",CP,
					leftdir,subdir,dir_contents[i].name,					
					rightdir,subdir,dir_contents[i].name);
				execute_command(string);
			}

			/* Otherwise files are in sync. */
			else
			{
			}

		}

		/* Both left and right are directories.	*/
		else if(dir_contents[i].left_type=='d' && 
			dir_contents[i].right_type=='d')
		{
			/* Process everything inside the directory. */
			sprintf(string,"%s%s/",subdir,dir_contents[i].name);
			walk_tree(string);
		}

		/* Left is a file and right does not exist. */
		else if(dir_contents[i].left_type=='f' && 
			dir_contents[i].right_type=='n')
		{
			/* Copy file to the right. */
			sprintf(string,"%s \"%s%s%s\" \"%s%s%s\"",CP,
				leftdir,subdir,dir_contents[i].name,
				rightdir,subdir,dir_contents[i].name);
			execute_command(string);
		}

		/* Left does not exist and right is a file. */
		else if(dir_contents[i].left_type=='n' && 
			dir_contents[i].right_type=='f')
		{
		}

		/* Left is a directory and right does not exist. */
		else if(dir_contents[i].left_type=='d' && 
			dir_contents[i].right_type=='n')
		{

			/* Create an empty directory to the right. */
			sprintf(string,"%s \"%s%s%s\"",MKDIR,
				rightdir,subdir,dir_contents[i].name);
			execute_command(string);
		
			/* Process everything inside the directory. */
			sprintf(string,"%s%s/",subdir,
				dir_contents[i].name);
			walk_tree(string);
		}

		/* Left does not exist and right is a directory. */
		else if(dir_contents[i].left_type=='n' && 
			dir_contents[i].right_type=='d')
		{
		}

		/* Left is a directory and right is a file. */
		else if(dir_contents[i].left_type=='d' && 
			dir_contents[i].right_type=='f')
		{

			/* Replace file to the right with an empty
			directory. */					
			sprintf(string,"%s \"%s%s%s\"",RM,
				rightdir,subdir,dir_contents[i].name);
			execute_command(string);
			sprintf(string,"%s \"%s%s%s\"",MKDIR,
				rightdir,subdir,dir_contents[i].name);
			execute_command(string);

			/* Now process everything inside the 
			directory. */
			sprintf(string,"%s%s/",
				subdir,dir_contents[i].name);
			walk_tree(string);

		}

		/* Left is a file and right is a directory. */
		else if(dir_contents[i].left_type=='f' && 
			dir_contents[i].right_type=='d')
		{
			/* Replace directory (tree) to the right
			with file to the left. */	
			sprintf(string,"%s \"%s%s%s\"",RMDIR,
				rightdir,subdir,dir_contents[i].name);
			execute_command(string);
			sprintf(string,"%s \"%s%s%s\" \"%s%s%s\"",CP,
				leftdir,subdir,dir_contents[i].name,
				rightdir,subdir,dir_contents[i].name);
			execute_command(string);
		}		

		/* We should not be able to get here. */
		else
		{
			fprintf(stderr,
				"%s: Panic. Backup aborted.\n",
				CMDNAME);
				exit(1);
		}
	}

	/* We can now release the memory allocated to store all the 
	filenames, and the memory allocated for the dir_contents
	structure and the memory allocated to store temporary
	strings. */
	for(i=0;dir_contents[i].name!=NULL;i++)free(dir_contents[i].name);
	free(dir_contents);
	free(string);
}



/***
*
* Execute a command and log it as well.
*
***/

void execute_command(char *command)
{
	int i,j,k;
	FILE *fp;

	char *processed_command;
	const char *escaped_single_quote=ESCAPED_SINGLE_QUOTE;

	/* Allocate memory for a processed copy of command plus an 
        terminating null character. We assume that worst case all
	characters are single quotes that need to be escaped. */
	processed_command=
		malloc((strlen(ESCAPED_SINGLE_QUOTE)+1)*strlen(command)+1);
	if(processed_command==NULL)
	{
		fprintf(stderr,"%s: Out of memory. Backup aborted.\n",
			CMDNAME);
		exit(1);
	}

	/* Copy command to processed_ command replacing slashes and quotes to 
	match the system we are compiled for. We start replacing after the 
	first space to cope with DOS command options, which are separated 
	with a slash from the command. */
	for(j=0,i=0;command[j]!=' ' && command[j]!='\0';j++,i++)
		processed_command[i]=command[j];
	for(;command[j]!='\0';j++)
	{
		if(command[j]=='\'')
		{
			for(k=0;escaped_single_quote[k]!='\0';k++)
				processed_command[i++]=escaped_single_quote[k];
		}
		else if(command[j]=='/')processed_command[i++]=SLASH;
		else if(command[j]=='"')processed_command[i++]=QUOTE;
                else processed_command[i++]=command[j];
	}
        processed_command[i]='\0';

	/* Log the command to be executed to the standard output. */
	fprintf(stdout,"%s\n",processed_command);

	/* If a logfile has been specified on the command line,
	log the command to be executed to that file as well. */
	if(logfile!=NULL)
	{
		fp=fopen(logfile,"at");
		if(fp==NULL)
		{
			fprintf(fp,"%s: Cannot open file \"%s\".\n",
				CMDNAME,logfile);
			exit(1);
		}
		fprintf(fp,"%s\n",processed_command);
		fclose(fp);
	}

	/* Execute the command. Exit this program if the command 
	exits with an error. */
	if(system(processed_command))
	{	
		fprintf(stderr,"%s: Backup aborted.\n",CMDNAME);
		exit(1);
	}

        /* Free memory for processed copy of comand again. */
        free(processed_command);
}



/***
*
* Check if file time t1 is greater t2. If ignore_one_hour is true,
* a time difference of 3600 s is considered to be no time difference.
* Accuracy of time is +/- 3 s.
*
***/

int gt_time(FILETIME t1,FILETIME t2,int ignore_one_hour)
{
	unsigned long int high1;
	unsigned long int low1;
	unsigned long int high2;
	unsigned long int low2;

	high1=t1.dwHighDateTime;
	low1=t1.dwLowDateTime;
	high2=t2.dwHighDateTime;
	low2=t2.dwLowDateTime;

	/* This means t1 < t2. */
	if( high1 < high2 )return 0;

	if( high1 == high2 )
	{
		/* This means t1 < t2. */
		if( low1 < low2 )return 0;

		/* This means t1 = t2 +/- 3 s. */
		if( ( low1 >= low2 && low1-low2 <= 30000000 ) || 
			( low2 >= low1 && low2-low1 <= 30000000 ) )return 0;
	}

	if( high1-high2 == 1)
	{
		/* This means t1 = t2 +/- 3 s. */
		if( low1 < low2 &&
			low1-low2 <= 30000000 )return 0;
	}

	if( ignore_one_hour && high1-high2 == 8)
	{
		/* This means t1 = t2 + 3600 s +/- 3 s,
		while we are ignoring one hour. */
		if( low2 < low1 &&
			low1-low2 <= 1670261632 &&
			low1-low2 >= 1610261632 )return 0;
	}

	if( ignore_one_hour && high1-high2 == 9)
	{
		/* This means t1 = t2 + 3600 s +/- 3 s,
		while we are ignoring one hour. */
		if( low1 < low2 &&
			low1-low2 <= 1670261632 &&
			low1-low2 >= 1610261632 )return 0;
	}

	/* This means t2 > t1. */
	return 1;

}




/***
*
* Check if entry is in database.
*
***/

int check_database(char *database,char *dir,char *entry)
{
	char *string;

	/* Allocate memory for a copy of dir plus a leading and trailing
	newline plus an terminating null character. */
	string=malloc(strlen(dir)+strlen(entry)+3);
	if(string==NULL)
	{
		fprintf(stderr,"%s: Out of memory. Backup aborted.\n",
			CMDNAME);
		exit(1);
	}

	/* Compose search string. */
	sprintf(string,"\n%s%s\n",dir,entry);

	/* Find string in database. Before returning the memory allocated
	to string must be freed again. */
	if(strstr(database,string)!=NULL)
	{
		free(string);
		return 1;
	}
	else 
	{
		free(string);
		return 0;
	}
}




/***
*
* Check if directory exists.
*
***/

int dir_exists(char *dir)
{
	char *string;
	int i;

	struct stat fileinfo;

	/* Allocate memory for a copy of dir. */
	string=malloc(strlen(dir) + 2 );
	if(string==NULL)
	{
		fprintf(stderr,"%s: Out of memory. Backup aborted.\n",
			CMDNAME);
		exit(1);
	}

	/* Copy dir to string. */
	for(i=0;dir[i]!='\0';i++)
	{
		string[i]=dir[i];
	}
	string[i]='\0';

	/* Remove any trailing slash unless this slash is the only character,
	or it is preceded by another slash or a colon. */
	if(i>=2 && (string[i-1]=='/' || string[i-1]=='\\') &&
		string[i-2]!='/' && string[i-2]!='\\' && string[i-2]!=':')
	{
		string[i-1]='\0';
	}
	
	/* Get info about dir. If we are unable to get info, we know that 
	dir does not exist. We shall not forget to free the memory
	allocated for string. */
	if(stat(string,&fileinfo)!=0)
	{
		free(string);
		return 0;
	}

	/* Depending on the file mode it is a directory or not. */
	if(fileinfo.st_mode & S_IFDIR)return 1;
	else return 0;

}
		
