/**
*
* Synchronize two directories.
*
**/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <windows.h>
#include <sys/types.h>
#include <sys/stat.h>

#define CMDNAME 		"syncdir"
#define VERSION 		"4.4"
#define COPYRIGHT		"copyright (C) 2009-2024 Frits Wiarda"
#define HELP			"Synchronize two directories.\n"	\
				"\n"					\
				"Usage: syncdir [-a] leftdir rightdir [logfile]\n"\
				"\n"\
				"Unless the -a option is specified, files with a name starting with\n"\
				"a dot, hidden files and system files are ignored.\n"
#define USAGE			"Usage: syncdir [-a] leftdir rightdir [logfile]\n"

#define SLASH			'\\'
#define QUOTE			'"'
#define ESCAPED_SINGLE_QUOTE    "'"
#define MKDIR			"mkdir"
#define RMDIR			"rm -rf"
#define CP			"cp -pf"
#define RM			"rm -f"

#define DATABASE_FILE		".syncdir"

#define DATABASE_SIZE		20000000
#define DATABASE_BLOCK_SIZE	65536
#define DIR_SIZE 		9192
#define FILENAME_SIZE		512
#define PARAMETER_SIZE		500

char leftdir[PARAMETER_SIZE];
char rightdir[PARAMETER_SIZE];
char *logfile;
int option_all;

char old_database[DATABASE_SIZE];
char new_database[DATABASE_SIZE];



void walk_tree(char *);
void execute_command(char *);
int check_database(char *,char *,char *);
int gt_time(FILETIME,FILETIME,int);
int dir_exists(char *);



int main(int argc, char *argv[])
{

	FILE *fp;

	char filename[FILENAME_SIZE];
	int arg_counter,non_option_arg_counter;
	int i,c;

	/* If first and only argument is "-?", "/?" or "--help",
	print copyright notice and help information. */
	if(argc==2 && (argv[1][0]=='-'||argv[1][0]=='/') && 
		argv[1][1]=='?' && argv[1][2]=='\0' || argc==2 && 
		argv[1][0]=='-' && argv[1][1]=='-' && 
		argv[1][2]=='h' && argv[1][3]=='e' && 
		argv[1][4]=='l' && argv[1][5]=='p' && 
		argv[1][6]=='\0')
	{
		fprintf(stdout,"%s %s   %s\n\n%s",
			CMDNAME,VERSION,COPYRIGHT,HELP);
		return 0;
	}

	/* Print usage message if parameters are not OK. */
	non_option_arg_counter=0;
	for(arg_counter=1; arg_counter<argc; arg_counter++)
	{
		if(argv[arg_counter][0]=='-')
		{
			if(strcmp(argv[arg_counter],"-a")!=0)
			{
				fprintf(stderr,"%s",USAGE);
				exit(1);
			}
		}
		else non_option_arg_counter++;
	}
	if(non_option_arg_counter<2 || non_option_arg_counter>3)
	{
		fprintf(stderr,"%s",USAGE);
		exit(1);
	}

	/* Evaluate the command line options. */
	option_all=0;
	for(arg_counter=1; arg_counter<argc; arg_counter++)
	{
		if(argv[arg_counter][0]=='-')
		{
			for(i=1; argv[arg_counter][i]!='\0'; i++)
			{
				if(argv[arg_counter][i]=='a')
				{
					option_all=1;
				}
			}
		}
	}

	/* Find the first non-option argument. */
	for(arg_counter=1; arg_counter<argc; arg_counter++)
	{
		if(argv[arg_counter][0]!='-')break;
	}

	/* Copy leftdir from command line, test if it exist,
	change backslashes into slashes and append a slash if 
	it is not ending with a slash yet. */
	if(!dir_exists(argv[arg_counter]))
	{
		fprintf(stderr,"%s: '%s' is not a directory.\n",CMDNAME,argv[arg_counter]);
		exit(1);
	}
	for(i=0;argv[arg_counter][i]!='\0' && i<PARAMETER_SIZE-2;i++)
	{
		leftdir[i]=argv[arg_counter][i];
		if(leftdir[i]=='\\')leftdir[i]='/';
	}
	if(leftdir[i-1]!='/')leftdir[i++]='/';
	leftdir[i]='\0';

	/* Find the second non-option argument. */
	for(arg_counter++; arg_counter<argc; arg_counter++)
	{
		if(argv[arg_counter][0]!='-')break;
	}

	/* Copy rightdir from command line, test if it exist,
	change backslashes into slashes and append a slash if 
	it is not ending with a slash yet. */
	if(!dir_exists(argv[arg_counter]))
	{
		fprintf(stderr,
			"%s: '%s' is not a directory.\n",CMDNAME,argv[arg_counter]);
		exit(1);
	}
	for(i=0;argv[arg_counter][i]!='\0' && i<PARAMETER_SIZE-2;i++)
	{
		rightdir[i]=argv[arg_counter][i];
		if(rightdir[i]=='\\')rightdir[i]='/';
	}
	if(rightdir[i-1]!='/')rightdir[i++]='/';
	rightdir[i]='\0';

	/*  Copy the optional logfile from the command line and create an 
	empty one (when it has been specified). */
	if(non_option_arg_counter==3)
	{
		/* Find the third non-option argument. */
		for(arg_counter++; arg_counter<argc; arg_counter++)
		{
			if(argv[arg_counter][0]!='-')break;
		}

		logfile=argv[arg_counter];
		fp=fopen(logfile,"wt");
		if(fp==NULL)
		{
			fprintf(stderr,"%s: Cannot open file '%s'.\n",
				CMDNAME,logfile);
			exit(1);
		}
		fclose(fp);
	}
	else
	{
		logfile=NULL;
	}

	/* Read database from leftdir, but only if it exists in the 
	rightdir as well. (If it would not exist in the rightdir a 
	typo in rightdir could result in the whole tree being deleted.) */
	sprintf(filename,"%s%s",rightdir,DATABASE_FILE);
	fp=fopen(filename,"rt");
	if(fp==NULL)
	{
		old_database[0]='\0';
	}
	else
	{
		fclose(fp);
		sprintf(filename,"%s%s",leftdir,DATABASE_FILE);
		fp=fopen(filename,"rt");
		if(fp==NULL)
		{
			old_database[0]='\0';
		}
		else
		{
			old_database[0]='\n';
			setvbuf(fp,NULL,_IOFBF,DATABASE_BLOCK_SIZE);
			for(i=1;;i++)
			{		
				c=getc(fp);
				while(c=='\r')c=getc(fp);
				if(c==EOF)break;
				if(i==DATABASE_SIZE-1)
				{
					fprintf(stderr,
						"%s: Database overflow.\n",
						CMDNAME);
					exit(1);
				}
				old_database[i]=(char)c;	
			}
			old_database[i]='\0';
		}
	}

	/* Initialize new database. */
	new_database[0]='\0';

	/* Walk through the directory tree. */
	walk_tree("");

	/* Write database to leftdir. */
	sprintf(filename,"%s%s",leftdir,DATABASE_FILE);
	fp=fopen(filename,"wt");
	if(fp==NULL)
	{
		fprintf(stderr,"%s: Cannot open file '%s'.\n",
			CMDNAME,filename);
		exit(1);
	}
	setvbuf(fp,NULL,_IOFBF,DATABASE_BLOCK_SIZE);
	for(i=0;new_database[i]!='\0';i++)
	{
		putc(new_database[i],fp);
	}
	fclose(fp);
	
	/* Write database to rightdir. */
	sprintf(filename,"%s%s",rightdir,DATABASE_FILE);
	fp=fopen(filename,"wt");
	if(fp==NULL)
	{
		fprintf(stderr,"%s: Cannot open file '%s'.\n",
			CMDNAME,filename);
		exit(1);
	}
	setvbuf(fp,NULL,_IOFBF,DATABASE_BLOCK_SIZE);
	for(i=0;new_database[i]!='\0';i++)
	{
		putc(new_database[i],fp);
	}
	fclose(fp);

	/* Ready. */
	return 0;

}



/***
*
* Walk through the directory tree.
*
***/

void walk_tree(char *subdir)
{

	struct dir_contents {
		char *name;
		char left_type;
		FILETIME left_time;
		long int left_size;
		char right_type;
		FILETIME right_time;
		long int right_size;
	} *dir_contents, tmp_dir_contents;

	int dir_contents_sorted;

	char *string;

	int i;
	int j;

	WIN32_FIND_DATA fileinfo;
	HANDLE h;

	/* Allocate memory for temporary strings. The maximum size needed
	is a command plus spaces and quotes plus a left pathname plus a
	right pathname. */
	string=malloc( 80 + 
		strlen(leftdir) + strlen(subdir) + FILENAME_SIZE + 
		strlen(rightdir) + strlen(subdir) + FILENAME_SIZE );
	if(string==NULL)
	{
		fprintf(stderr,"%s: Out of memory. Synchronization aborted.\n",
			CMDNAME);
		exit(1);
	}

	/* Allocate memory for directory contents. */
	dir_contents=malloc(sizeof(struct dir_contents)*DIR_SIZE);
	if(dir_contents==NULL)
	{
		fprintf(stderr,"%s: Out of memory. Synchronization aborted.\n",
			CMDNAME);
		exit(1);
	}

	/* Set the terminating NULL to dir_contents. */
	dir_contents[0].name=NULL;

	/* Read directory to the left, and copy the information into
	dir_contents. */
	sprintf(string,"%s%s*",leftdir,subdir);
	h=FindFirstFileA((LPCSTR)string,&fileinfo);
	if(h!=INVALID_HANDLE_VALUE)for(i=0;;)
	{
		/* We always skip ".", ".." and the database file. */
		if(strcmp(fileinfo.cFileName,".")==0 || 
			strcmp(fileinfo.cFileName,"..")==0 || 
			strcmp(fileinfo.cFileName,DATABASE_FILE)==0 )
		{
			if(FindNextFileA(h,&fileinfo)==0)break;
			else continue;
		}

		/* We skip hidden and system files, and files with a name
		starting with a dot, unless the "-a" option has been 
		specified. */
		if(!option_all)
		{
			/* Hidden and system files. */
			if(fileinfo.dwFileAttributes & 
				FILE_ATTRIBUTE_HIDDEN ||
				fileinfo.dwFileAttributes & 
				FILE_ATTRIBUTE_SYSTEM)
			{
				if(FindNextFileA(h,&fileinfo)==0)break;
				else continue;
			}

			/* Files with a name starting with a dot. */
			if(fileinfo.cFileName[0]=='.')
			{
				if(FindNextFileA(h,&fileinfo)==0)break;
				else continue;
			}
		}

		/* Add entry to dir_contents. */
		dir_contents[i].name=malloc(strlen(fileinfo.cFileName)+1);
		if(dir_contents[i].name==NULL)
		{
			fprintf(stderr,
				"%s: Out of memory. Synchronization aborted.\n",
				CMDNAME);
			exit(1);
		}
		strcpy(dir_contents[i].name,fileinfo.cFileName);
		if(fileinfo.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
		{
			dir_contents[i].left_type='d';
		}
		else 
		{
			dir_contents[i].left_type='f';
		}
		dir_contents[i].left_time=fileinfo.ftLastWriteTime;
      		dir_contents[i].left_size=fileinfo.nFileSizeHigh * MAXDWORD +
			fileinfo.nFileSizeLow;
		dir_contents[i].right_type='n';

		/* Increase counter and add the terminating NULL. */
		i++;
		dir_contents[i].name=NULL;

		/* Get next entry. */
		if(FindNextFileA(h,&fileinfo)==0)break;

		/* Check that there is space left in the array dir_contents. */
		if(i>=DIR_SIZE-1)
		{
			fprintf(stderr,
				"%s: To many entries in directory '%s%s'. Synchronization aborted.\n",
				CMDNAME,leftdir,subdir);
			exit(1);
		}
	}
	if(h!=INVALID_HANDLE_VALUE)FindClose(h);

	/* Read directory to the right, and copy the information into
	dir_contents. */
	sprintf(string,"%s%s*",rightdir,subdir);
	h=FindFirstFileA((LPCSTR)string,&fileinfo);
	if(h!=INVALID_HANDLE_VALUE)for(;;)
	{

		/* We always skip ".", ".." and the database file. */
		if(strcmp(fileinfo.cFileName,".")==0 || 
			strcmp(fileinfo.cFileName,"..")==0 || 
			strcmp(fileinfo.cFileName,DATABASE_FILE)==0 )
		{
			if(FindNextFileA(h,&fileinfo)==0)break;
			else continue;
		}

		/* We skip hidden and system files, and files with a name
		starting with a dot, unless the "-a" option has been 
		specified. */
		if(!option_all)
		{
			/* Hidden and system files. */
			if(fileinfo.dwFileAttributes & 
				FILE_ATTRIBUTE_HIDDEN ||
				fileinfo.dwFileAttributes & 
				FILE_ATTRIBUTE_SYSTEM)
			{
				if(FindNextFileA(h,&fileinfo)==0)break;
				else continue;
			}

			/* Files with a name starting with a dot. */
			if(fileinfo.cFileName[0]=='.')
			{
				if(FindNextFileA(h,&fileinfo)==0)break;
				else continue;
			}
		}

		/* Try to find filename in dir_contents structure. */
		for(j=0;dir_contents[j].name!=NULL;j++)
		{
			if(strcmp(dir_contents[j].name,
				fileinfo.cFileName)==0)break;
		}

		/* If filename found we only have to add the file data to
		the existing entry. */		
		if(dir_contents[j].name!=NULL)
		{
			/* Add file data to existing entry. */
			if(fileinfo.dwFileAttributes & 
				FILE_ATTRIBUTE_DIRECTORY)
			{
				dir_contents[j].right_type='d';
			}
			else 
			{
				dir_contents[j].right_type='f';
			}			
			dir_contents[j].right_time=fileinfo.ftLastWriteTime;
			dir_contents[j].right_size=
				fileinfo.nFileSizeHigh * MAXDWORD +
				fileinfo.nFileSizeLow;
		}

		/* Otherwise we have to create a new entry. */
		else
		{
			/* Add entry to dir_contents. */
			dir_contents[i].name=
				malloc(strlen(fileinfo.cFileName)+1);
			if(dir_contents[i].name==NULL)
			{
				fprintf(stderr,
					"%s: Out of memory. Synchronization aborted.\n",
					CMDNAME);
				exit(1);
			}
			strcpy(dir_contents[i].name,fileinfo.cFileName);
			dir_contents[i].left_type='n';
			if(fileinfo.dwFileAttributes & 
				FILE_ATTRIBUTE_DIRECTORY)
			{
				dir_contents[i].right_type='d';
			}
			else 
			{
				dir_contents[i].right_type='f';
			}
			dir_contents[i].right_time=fileinfo.ftLastWriteTime;
			dir_contents[i].right_size=
				fileinfo.nFileSizeHigh * MAXDWORD +
				fileinfo.nFileSizeLow;

			/* Increase counter and add the terminating NULL. */
			i++;
			dir_contents[i].name=NULL;
		}

		/* Get next entry. */
		if(FindNextFileA(h,&fileinfo)==0)break;

		/* Check that there is space left in the array dir_contents. */
		if(i>=DIR_SIZE-1)
		{
			fprintf(stderr,
				"%s: To many entries in directory '%s%s'. Synchronization aborted.\n",
				CMDNAME,leftdir,subdir);
			exit(1);
		}
	}
	if(h!=INVALID_HANDLE_VALUE)FindClose(h);

	/* Sort the array dir_contents when there is more then
	one entry in it. */
	if(dir_contents[0].name!=NULL && dir_contents[1].name!=NULL)
	{
		do
		{
			dir_contents_sorted=1;
			for(i=1;dir_contents[i].name!=NULL;i++)
			{
				if(strcmp(dir_contents[i-1].name,
					dir_contents[i].name)>0)
				{
					dir_contents_sorted=0;
						tmp_dir_contents=
							dir_contents[i-1];
						dir_contents[i-1]=
							dir_contents[i];
						dir_contents[i]=
							tmp_dir_contents;
				}
			}
		}
		while(!dir_contents_sorted);
	}

	/* Now loop through the array dir_contents. Do the
	removal part of the synchronization first. Remove before
	copy ensures correct functioning on filesystems that
	consider lowercase and uppercase names to be identical. */
	for(i=0;dir_contents[i].name!=NULL;i++)
	{

		/* Both left and right are files. */
		if(dir_contents[i].left_type=='f' && 
			dir_contents[i].right_type=='f')
		{
			/* This will be done later. */
		}

		/* Both left and right are directories.	*/
		else if(dir_contents[i].left_type=='d' && 
			dir_contents[i].right_type=='d')
		{
			/* This will be done later. */
		}

		/* Left is a file and right does not exist. */
		else if(dir_contents[i].left_type=='f' && 
			dir_contents[i].right_type=='n')
		{
			/* Check in database if entry did exist before. */
			if(check_database(old_database,subdir,
				dir_contents[i].name))
			{
				/* Check that parent directory of deleted 
				file is still readable. This is done to 
				prevent a whole tree being deleted if a 
				network drive gets disconnected while 
				synchronizing. */
				sprintf(string,"%s%s",rightdir,subdir);
				if(!dir_exists(string))
				{
					fprintf(stderr,
						"%s: Directory '%s' is not readable. Synchronization aborted.\n",
						CMDNAME,string);
					exit(1);
				}
			
				/* Delete file to the left. */
				sprintf(string,"%s \"%s%s%s\"",RM,
					leftdir,subdir,dir_contents[i].name);
				execute_command(string);
			}

			/* If entry is not in the database it has recently 
			been created. */
			else
			{
				/* This will be done later. */
			}
		}

		/* Left does not exist and right is a file. */
		else if(dir_contents[i].left_type=='n' && 
			dir_contents[i].right_type=='f')
		{
			/* Check in database if entry did exist before. */
			if(check_database(old_database,subdir,
				dir_contents[i].name))
			{	
				/* Check that parent directory of deleted 
				file is still readable. This is done to 
				prevent a whole tree being deleted if a 
				network drive gets disconnected while 
				synchronizing. */
				sprintf(string,"%s%s",leftdir,subdir);
				if(!dir_exists(string))
				{
					fprintf(stderr,
						"%s: Directory '%s' is not readable. Synchronization aborted.\n",
						CMDNAME,string);
					exit(1);
				}

				/* Delete file to the right. */
				sprintf(string,"%s \"%s%s%s\"",RM,
					rightdir,subdir,dir_contents[i].name);
				execute_command(string);
			}

			/* If entry is not in the database it has recently 
			been created. */
			else
			{
				/* This will be done later. */
			}
		}

		/* Left is a directory and right does not exist. */
		else if(dir_contents[i].left_type=='d' && 
			dir_contents[i].right_type=='n')
		{
			/* Check in database if entry did exist before. */
			if(check_database(old_database,subdir,
				dir_contents[i].name))
			{
				/* Check that parent directory of deleted 
				file is still readable. This is done to 
				prevent a whole tree being deleted if a 
				network drive gets disconnected while 
				synchronizing. */
				sprintf(string,"%s%s",rightdir,subdir);
				if(!dir_exists(string))
				{
					fprintf(stderr,
						"%s: Directory '%s' is not readable. Synchronization aborted.\n",
						CMDNAME,string);
					exit(1);
				}

				/* Delete directory (tree) to the left. */
				sprintf(string,"%s \"%s%s%s\"",RMDIR,
					leftdir,subdir,dir_contents[i].name);
				execute_command(string);
			}

			/* If entry is not in the database it has recently 
			been created. */
			else
			{
				/* This will be done later. */
			}
		}

		/* Left does not exist and right is a directory. */
		else if(dir_contents[i].left_type=='n' && 
			dir_contents[i].right_type=='d')
		{
			/* Check in database if entry did exist before. */
			if(check_database(old_database,subdir,
				dir_contents[i].name))
			{
				/* Check that parent directory of deleted 
				file is still readable. This is done to 
				prevent a whole tree being deleted if a 
				network drive gets disconnected while 
				synchronizing. */
				sprintf(string,"%s%s",leftdir,subdir);
				if(!dir_exists(string))
				{
					fprintf(stderr,
						"%s: Directory '%s' is not readable. Synchronization aborted.\n",
						CMDNAME,string);
					exit(1);
				}

				/* Delete directory (tree) to the right. */	
				sprintf(string,"%s \"%s%s%s\"",RMDIR,
					rightdir,subdir,dir_contents[i].name);
				execute_command(string);
			}

			/* If entry is not in the database it has recently 
			been created. */
			else
			{
				/* This will be done later. */
			}
		}

		/* Left is a directory and right is a file. */
		else if(dir_contents[i].left_type=='d' && 
			dir_contents[i].right_type=='f')
		{
			/* This will be done later. */
		}

		/* Left is a file and right is a directory. */
		else if(dir_contents[i].left_type=='f' && 
			dir_contents[i].right_type=='d')
		{
			/* This will be done later. */
		}		

		/* We should not be able to get here. */
		else
		{
			fprintf(stderr,
				"%s: Panic. Synchronization aborted.\n",
				CMDNAME);
				exit(1);
		}
	}

	/* Now loop through the array dir_contents again and do 
	the rest of the synchronization. */
	for(i=0;dir_contents[i].name!=NULL;i++)
	{

		/* Both left and right are files. */
		if(dir_contents[i].left_type=='f' && 
			dir_contents[i].right_type=='f')
		{

			/* Check if left is newer then right. */
			if( gt_time(dir_contents[i].left_time,
				dir_contents[i].right_time,
				dir_contents[i].left_size==
				dir_contents[i].right_size) )
			{
				/* Update file to the right. */
				sprintf(string,"%s \"%s%s%s\" \"%s%s%s\"",CP,
					leftdir,subdir,dir_contents[i].name,					
					rightdir,subdir,dir_contents[i].name);
				execute_command(string);
			}

			/* Check if right is newer then left. */
			else if( gt_time(dir_contents[i].right_time,
				dir_contents[i].left_time,
				dir_contents[i].left_size==
				dir_contents[i].right_size) )
			{
				/* Update file to the left. */
				sprintf(string,"%s \"%s%s%s\" \"%s%s%s\"",CP,
					rightdir,subdir,dir_contents[i].name,
					leftdir,subdir,dir_contents[i].name);
				execute_command(string);
			}

			/* Otherwise files are in sync. */
			else
			{
			}

			/* Add entry to database. */
			strcat(new_database,subdir);
			strcat(new_database,dir_contents[i].name);
			strcat(new_database,"\n");

		}

		/* Both left and right are directories.	*/
		else if(dir_contents[i].left_type=='d' && 
			dir_contents[i].right_type=='d')
		{
			/* Add entry to database. */
			strcat(new_database,subdir);
			strcat(new_database,dir_contents[i].name);
			strcat(new_database,"\n");

			/* Process everything inside the directory. */
			sprintf(string,"%s%s/",subdir,dir_contents[i].name);
			walk_tree(string);
		}

		/* Left is a file and right does not exist. */
		else if(dir_contents[i].left_type=='f' && 
			dir_contents[i].right_type=='n')
		{
			/* Check in database if entry did exist before. */
			if(check_database(old_database,subdir,
				dir_contents[i].name))
			{
				/* Removal has already been done. */
			}

			/* If entry is not in the database it has recently 
			been created. */
			else
			{
				/* Copy file to the right. */
				sprintf(string,"%s \"%s%s%s\" \"%s%s%s\"",CP,
					leftdir,subdir,dir_contents[i].name,
					rightdir,subdir,dir_contents[i].name);
				execute_command(string);

				/* Add entry to database. */
				strcat(new_database,subdir);
				strcat(new_database,dir_contents[i].name);
				strcat(new_database,"\n");
			}
		}

		/* Left does not exist and right is a file. */
		else if(dir_contents[i].left_type=='n' && 
			dir_contents[i].right_type=='f')
		{
			/* Check in database if entry did exist before. */
			if(check_database(old_database,subdir,
				dir_contents[i].name))
			{	
				/* Removal has already been done. */
			}

			/* If entry is not in the database it has recently 
			been created. */
			else
			{
				/* Copy file to the left. */
				sprintf(string,"%s \"%s%s%s\" \"%s%s%s\"",CP,
					rightdir,subdir,dir_contents[i].name,
					leftdir,subdir,dir_contents[i].name);
				execute_command(string);


				/* Add entry to database. */
				strcat(new_database,subdir);
				strcat(new_database,dir_contents[i].name);
				strcat(new_database,"\n");
			}
		}

		/* Left is a directory and right does not exist. */
		else if(dir_contents[i].left_type=='d' && 
			dir_contents[i].right_type=='n')
		{
			/* Check in database if entry did exist before. */
			if(check_database(old_database,subdir,
				dir_contents[i].name))
			{
				/* Removal has already been done. */
			}

			/* If entry is not in the database it has recently 
			been created. */
			else
			{
				/* Create an empty directory to the right. */
				sprintf(string,"%s \"%s%s%s\"",MKDIR,
					rightdir,subdir,dir_contents[i].name);
				execute_command(string);
		
				/* Add entry to database. */
				strcat(new_database,subdir);
				strcat(new_database,dir_contents[i].name);
				strcat(new_database,"\n");
	
				/* Process everything inside the directory. */
				sprintf(string,"%s%s/",subdir,
					dir_contents[i].name);
				walk_tree(string);
			}
		}

		/* Left does not exist and right is a directory. */
		else if(dir_contents[i].left_type=='n' && 
			dir_contents[i].right_type=='d')
		{
			/* Check in database if entry did exist before. */
			if(check_database(old_database,subdir,
				dir_contents[i].name))
			{
				/* Removal has already been done. */
			}

			/* If entry is not in the database it has recently 
			been created. */
			else
			{

				/* Create empty directory to the left. */					
				sprintf(string,"%s \"%s%s%s\"",MKDIR,
					leftdir,subdir,dir_contents[i].name);
				execute_command(string);

				/* Add entry to database. */
				strcat(new_database,subdir);
				strcat(new_database,dir_contents[i].name);
				strcat(new_database,"\n");
	
				/* Process everything inside the directory. */
				sprintf(string,"%s%s/",subdir,dir_contents[i].name);
				walk_tree(string);

			}
		}

		/* Left is a directory and right is a file. */
		else if(dir_contents[i].left_type=='d' && 
			dir_contents[i].right_type=='f')
		{
			/* Check if left is newer then right. */
			if( gt_time(dir_contents[i].left_time,
				dir_contents[i].right_time,0) )
			{

				/* Check that parent directory to the left
				is still readable. This is done to prevent 
				erroneous deletes when a network drive 
				gets disconnected while synchronizing. */
				sprintf(string,"%s%s",leftdir,subdir);
				if(!dir_exists(string))
				{
					fprintf(stderr,
						"%s: Directory '%s' is not readable. Synchronization aborted.\n",
						CMDNAME,string);
					exit(1);
				}

				/* Replace file to the right with an empty
				directory. */					
				sprintf(string,"%s \"%s%s%s\"",RM,
					rightdir,subdir,dir_contents[i].name);
				execute_command(string);
				sprintf(string,"%s \"%s%s%s\"",MKDIR,
					rightdir,subdir,dir_contents[i].name);
				execute_command(string);

				/* Add entry to database. */
				strcat(new_database,subdir);
				strcat(new_database,dir_contents[i].name);
				strcat(new_database,"\n");		

				/* Now process everything inside the 
				directory. */
				sprintf(string,"%s%s/",
					subdir,dir_contents[i].name);
				walk_tree(string);
			}

			/* Check if right is newer then left. */
			else if( gt_time(dir_contents[i].right_time,
				dir_contents[i].left_time,0) )
			{

				/* Check that parent directory to the right
				is still readable. This is done to prevent 
				erroneous deletes when a network drive 
				gets disconnected while synchronizing. */
				sprintf(string,"%s%s",rightdir,subdir);
				if(!dir_exists(string))
				{
					fprintf(stderr,
						"%s: Directory '%s' is not readable. Synchronization aborted.\n",
						CMDNAME,string);
					exit(1);
				}

				/* Replace directory (tree) to the left
				with file to the right. */			
				sprintf(string,"%s \"%s%s%s\"",RMDIR,
					leftdir,subdir,dir_contents[i].name);
				execute_command(string);
				sprintf(string,"%s \"%s%s%s\" \"%s%s%s\"",CP,
					rightdir,subdir,dir_contents[i].name,
					leftdir,subdir,dir_contents[i].name);
				execute_command(string);

				/* Add entry to database. */
				strcat(new_database,subdir);
				strcat(new_database,dir_contents[i].name);
				strcat(new_database,"\n");
			}
			else
			{
				/* This should not happen. */
				fprintf(stderr,
					"%s: Directory '%s%s%s' is equally old as file '%s%s%s'. Synchronization aborted.\n",
					CMDNAME,
					leftdir,subdir,dir_contents[i].name,
					rightdir,subdir,dir_contents[i].name);
				exit(1);
			}
		}

		/* Left is a file and right is a directory. */
		else if(dir_contents[i].left_type=='f' && 
			dir_contents[i].right_type=='d')
		{
			/* Check if left is newer then right. */
			if( gt_time(dir_contents[i].left_time,
				dir_contents[i].right_time,0) )
			{
				/* Check that parent directory to the left
				is still readable. This is done to prevent 
				erroneous deletes when a network drive 
				gets disconnected while synchronizing. */
				sprintf(string,"%s%s",leftdir,subdir);
				if(!dir_exists(string))
				{
					fprintf(stderr,
						"%s: Directory '%s' is not readable. Synchronization aborted.\n",
						CMDNAME,string);
					exit(1);
				}

				/* Replace directory (tree) to the right
				with file to the left. */	
				sprintf(string,"%s \"%s%s%s\"",RMDIR,
					rightdir,subdir,dir_contents[i].name);
				execute_command(string);
				sprintf(string,"%s \"%s%s%s\" \"%s%s%s\"",CP,
					leftdir,subdir,dir_contents[i].name,
					rightdir,subdir,dir_contents[i].name);
				execute_command(string);

				/* Add entry to database. */
				strcat(new_database,subdir);
				strcat(new_database,dir_contents[i].name);
				strcat(new_database,"\n");
			}

			/* Check if right is newer then left. */
			else if( gt_time(dir_contents[i].right_time,
				dir_contents[i].left_time,0) )
			{
				/* Check that parent directory to the left
				is still readable. This is done to prevent 
				erroneous deletes when a network drive 
				gets disconnected while synchronizing. */
				sprintf(string,"%s%s",rightdir,subdir);
				if(!dir_exists(string))
				{
					fprintf(stderr,
						"%s: Directory '%s' is not readable. Synchronization aborted.\n",
						CMDNAME,string);
					exit(1);
				}

				/* Replace file to the left with an empty
				directory. */					
				sprintf(string,"%s \"%s%s%s\"",RM,
					leftdir,subdir,dir_contents[i].name);
				execute_command(string);
				sprintf(string,"%s \"%s%s%s\"",MKDIR,
					leftdir,subdir,dir_contents[i].name);
				execute_command(string);

				/* Add entry to database. */
				strcat(new_database,subdir);
				strcat(new_database,dir_contents[i].name);
				strcat(new_database,"\n");
	
				/* Process everything inside the directory. */
				sprintf(string,"%s%s/",
					subdir,dir_contents[i].name);
				walk_tree(string);

			}
			else
			{
				/* This should not happen. */
				fprintf(stderr,
					"%s: File '%s%s%s' is equally old as directory '%s%s%s'. Synchronization aborted.\n",
					CMDNAME,
					leftdir,subdir,dir_contents[i].name,
					rightdir,subdir,dir_contents[i].name);
				exit(1);
			}
		}		

		/* We should not be able to get here. */
		else
		{
			fprintf(stderr,
				"%s: Panic. Synchronization aborted.\n",
				CMDNAME);
				exit(1);
		}
	}

	/* We can now release the memory allocated to store all the 
	filenames, and the memory allocated for the dir_contents
	structure and the memory allocated to store temporary
	strings. */
	for(i=0;dir_contents[i].name!=NULL;i++)free(dir_contents[i].name);
	free(dir_contents);
	free(string);
}



/***
*
* Execute a command and log it as well.
*
***/

void execute_command(char *command)
{
	int i,j,k;
	FILE *fp;

	char *processed_command;
	const char *escaped_single_quote=ESCAPED_SINGLE_QUOTE;

	/* Allocate memory for a processed copy of command plus an 
        terminating null character. We assume that worst case all
	characters are single quotes that need to be escaped. */
	processed_command=
		malloc((strlen(ESCAPED_SINGLE_QUOTE)+1)*strlen(command)+1);
	if(processed_command==NULL)
	{
		fprintf(stderr,"%s: Out of memory. Synchronisation aborted.\n",
			CMDNAME);
		exit(1);
	}

	/* Copy command to processed_ command replacing slashes and quotes to 
	match the system we are compiled for. We start replacing after the 
	first space to cope with DOS command options, which are separated 
	with a slash from the command. */
	for(j=0,i=0;command[j]!=' ' && command[j]!='\0';j++,i++)
		processed_command[i]=command[j];
	for(;command[j]!='\0';j++)
	{
		if(command[j]=='\'')
		{
			for(k=0;escaped_single_quote[k]!='\0';k++)
				processed_command[i++]=escaped_single_quote[k];
		}
		else if(command[j]=='/')processed_command[i++]=SLASH;
		else if(command[j]=='"')processed_command[i++]=QUOTE;
                else processed_command[i++]=command[j];
	}
        processed_command[i]='\0';

	/* Log the command to be executed to the standard output. */
	fprintf(stdout,"%s\n",processed_command);

	/* If a logfile has been specified on the command line,
	log the command to be executed to that file as well. */
	if(logfile!=NULL)
	{
		fp=fopen(logfile,"at");
		if(fp==NULL)
		{
			fprintf(fp,"%s: Cannot open file \"%s\".\n",
				CMDNAME,logfile);
			exit(1);
		}
		fprintf(fp,"%s\n",processed_command);
		fclose(fp);
	}

	/* Execute the command. Exit this program if the command 
	exits with an error. */
	if(system(processed_command))
	{	
		fprintf(stderr,"%s: Synchronisation aborted.\n",CMDNAME);
		exit(1);
	}

        /* Free memory for processed copy of comand again. */
        free(processed_command);
}




/***
*
* Check if file time t1 is greater t2. If ignore_one_hour is true,
* a time difference of 3600 s is considered to be no time difference.
* Accuracy of time is +/- 3 s.
*
***/

int gt_time(FILETIME t1,FILETIME t2,int ignore_one_hour)
{
	unsigned long int high1;
	unsigned long int low1;
	unsigned long int high2;
	unsigned long int low2;

	high1=t1.dwHighDateTime;
	low1=t1.dwLowDateTime;
	high2=t2.dwHighDateTime;
	low2=t2.dwLowDateTime;

	/* This means t1 < t2. */
	if( high1 < high2 )return 0;

	if( high1 == high2 )
	{
		/* This means t1 < t2. */
		if( low1 < low2 )return 0;

		/* This means t1 = t2 +/- 3 s. */
		if( ( low1 >= low2 && low1-low2 <= 30000000 ) || 
			( low2 >= low1 && low2-low1 <= 30000000 ) )return 0;
	}

	if( high1-high2 == 1)
	{
		/* This means t1 = t2 +/- 3 s. */
		if( low1 < low2 &&
			low1-low2 <= 30000000 )return 0;
	}

	if( ignore_one_hour && high1-high2 == 8)
	{
		/* This means t1 = t2 + 3600 s +/- 3 s,
		while we are ignoring one hour. */
		if( low2 < low1 &&
			low1-low2 <= 1670261632 &&
			low1-low2 >= 1610261632 )return 0;
	}

	if( ignore_one_hour && high1-high2 == 9)
	{
		/* This means t1 = t2 + 3600 s +/- 3 s,
		while we are ignoring one hour. */
		if( low1 < low2 &&
			low1-low2 <= 1670261632 &&
			low1-low2 >= 1610261632 )return 0;
	}

	/* This means t2 > t1. */
	return 1;

}




/***
*
* Check if entry is in database.
*
***/

int check_database(char *database,char *dir,char *entry)
{
	char *string;

	/* Allocate memory for a copy of dir plus a leading and trailing
	newline plus a termination null character. */
	string=malloc(strlen(dir)+strlen(entry)+3);
	if(string==NULL)
	{
		fprintf(stderr,"%s: Out of memory. Synchronization aborted.\n",
			CMDNAME);
		exit(1);
	}

	/* Compose search string. */
	sprintf(string,"\n%s%s\n",dir,entry);

	/* Find string in database. Before returning the memory allocated
	to string must be freed again. */
	if(strstr(database,string)!=NULL)
	{
		free(string);
		return 1;
	}
	else 
	{
		free(string);
		return 0;
	}
}




/***
*
* Check if directory exists.
*
***/

int dir_exists(char *dir)
{
	char *string;
	int i;

	struct stat fileinfo;

	/* Allocate memory for a copy of dir. */
	string=malloc(strlen(dir) + 2 );
	if(string==NULL)
	{
		fprintf(stderr,"%s: Out of memory. Synchronization aborted.\n",
			CMDNAME);
		exit(1);
	}

	/* Copy dir to string. */
	for(i=0;dir[i]!='\0';i++)
	{
		string[i]=dir[i];
	}
	string[i]='\0';

	/* Remove any trailing slash unless this slash is the only character,
	or it is preceded by another slash or a colon. */
	if(i>=2 && (string[i-1]=='/' || string[i-1]=='\\') &&
		string[i-2]!='/' && string[i-2]!='\\' && string[i-2]!=':')
	{
		string[i-1]='\0';
	}
	
	/* Get info about dir. If we are unable to get info, we know that 
	dir does not exist. We shall not forget to free the memory
	allocated for string. */
	if(stat(string,&fileinfo)!=0)
	{
		free(string);
		return 0;
	}

	/* Depending on the file mode it is a directory or not. */
	if(fileinfo.st_mode & S_IFDIR)return 1;
	else return 0;

}
